import sys
from typing import Callable
from pythreepio.errors import TranslationMissing


class Placeholder(object):
    def __init__(self, key, value=None):
        self.key = key
        self.value = value


class Command(object):
    def __init__(
        self,
        function_name: str,
        args: list,
        kwargs: dict,
        attrs: list = [],
        placeholder_output: str = None,
        exec_fn: Callable = None,
    ):
        self.function_name = function_name
        self.args = args
        self.kwargs = kwargs
        self.attrs = attrs
        self.placeholder_output = placeholder_output
        self.exec_fn = exec_fn
        self.frameworks = {"tf": "tensorflow"}

    def execute_routine(self, store={}):
        if self.exec_fn is None:
            attrs = self.attrs.copy()
            translated_cmd = sys.modules[self.frameworks[attrs.pop(0)]]
            while len(attrs) > 0:
                translated_cmd = getattr(translated_cmd, attrs.pop(0))
            self.exec_fn = translated_cmd

        args = []
        for arg in self.args:
            if isinstance(arg, Placeholder):
                args.append(store[arg.key])
            else:
                args.append(arg)

        return self.exec_fn(*args, **self.kwargs)


def cmd_from_info(info, store: dict) -> Command:
    args = []
    kwargs = {}
    for arg in info["args"]:
        is_kwarg = arg["kwarg"]
        name = arg.get("placeholder_input", arg["name"])
        if is_kwarg is True:
            if name in store:
                kwargs[name] = store[name]
        else:
            if name in store:
                args.append(store[name])
            else:
                args.append(Placeholder(name))

    return Command(
        info["name"],
        args,
        kwargs,
        info.get("attrs", []),
        info.get("placeholder_output", None),
    )



import re

from typing import Tuple, List
from .utils import get_mapped_commands
from .errors import TranslationMissing
from .command import Command, cmd_from_info


class Threepio(object):
    def __init__(self, from_lang: str, to_lang: str, framework: object):
        self.commands = get_mapped_commands()
        self.from_lang = from_lang
        self.to_lang = to_lang
        self.framework = framework

    def _normalize_func_name(self, name: str, lang: str) -> str:
        alpha = re.compile("[^a-zA-Z]")
        return alpha.sub("", name).lower()

    def _order_args(
        self, cmd: Command, from_info: dict, to_info: dict
    ) -> Tuple[list, dict]:
        new_args = []
        new_kwargs = {}
        for i, arg in enumerate(cmd.args):
            from_arg = from_info["args"][i]
            to_arg_index = next(
                (
                    index
                    for index, d in enumerate(to_info["args"])
                    if d["name"] == from_arg.get(self.to_lang, None)
                ),
                None,
            )

            if to_arg_index is None:
                new_args.append(arg)
                continue

            new_args.insert(to_arg_index, arg)

        # If any kwargs are normal args, splice them in as well
        for k, v in cmd.kwargs.items():
            from_arg = [a for a in from_info["args"] if a["name"] == k][0]
            to_arg_index = next(
                (
                    index
                    for index, d in enumerate(to_info["args"])
                    if d["name"] == from_arg.get(self.to_lang, {})
                ),
                None,
            )

            if to_arg_index is None:
                new_kwargs[k] = v
                continue

            new_args.insert(to_arg_index, v)

        return new_args, new_kwargs

    def translate_multi(self, orig_cmd, commands_info):
        cmd_config = commands_info.pop(0)
        store = {}
        for i, arg in enumerate(orig_cmd.args):
            cmd_config["args"][i]["value"] = arg
            store[cmd_config["args"][i]["name"]] = arg

        new_cmds = [cmd_config]
        for from_info in commands_info:
            cmd = cmd_from_info(from_info, store)
            to_info = self.commands[self.to_lang][
                self._normalize_func_name(from_info.get(self.to_lang), self.to_lang)
            ][0]

            new_cmds.append(self.translate_command(cmd, from_info, to_info))
        return new_cmds

    def translate_command(self, cmd, from_command, to_command):
        attrs = to_command["attrs"][1:]
        translated_cmd = None
        args, kwargs = self._order_args(cmd, from_command, to_command)
        output = from_command.get("placeholder_output", None)

        return Command(
            to_command["name"],
            args,
            kwargs,
            attrs=to_command["attrs"],
            placeholder_output=output,
            exec_fn=translated_cmd,
        )

    def translate(self, cmd: Command, lookup_command: bool = False) -> List[Command]:
        from_info = self.commands[self.from_lang][
            self._normalize_func_name(cmd.function_name, self.from_lang)
        ]
        if len(from_info) > 1:
            return self.translate_multi(cmd, from_info)

        from_info = from_info[0]

        if from_info.get(self.to_lang, None) is None:
            raise TranslationMissing(cmd.function_name)

        to_info = self.commands[self.to_lang][
            self._normalize_func_name(from_info.get(self.to_lang), self.to_lang)
        ]

        return [self.translate_command(cmd, from_info, to_info[0])]



class TranslationMissing(Exception):
    def __init__(self, name):
        super().__init__(
            f"Translation for the command {name} is not currently supported"
        )


class NotTranslated(Exception):
    def __init__(self):
        super().__init__("Translation must be completed before executing")



import pytest
import tensorflow as tf

from pythreepio.threepio import Threepio
from tests.fixtures.torch import (
    abs as torch_abs,
    add,
    add2,
    matmul,
    mean,
    mul,
    div,
    eq,
    sum as torch_sum,
    argmax,
    t,
    sub,
    truediv,
    linear,
)


@pytest.fixture
def threepio():
    return Threepio("torch", "tf", tf)


def round(val, decimals=0):
    multiplier = tf.constant(10 ** decimals, dtype=val.dtype)
    return tf.round(val * multiplier) / multiplier


def check_answer(result, answer):
    assert tf.reduce_all(tf.equal(round(result, 3), round(answer, 3)))


def process_tests(command, threepio):
    for i, input in enumerate(command["inputs"]):
        translations = threepio.translate(input, lookup_command=True)
        for j, translation in enumerate(translations):
            result = translation.execute_routine()
            answer = command["answers"][i][j]
            check_answer(result, answer)


def test_translates_abs(threepio):
    process_tests(torch_abs, threepio)


@pytest.mark.skip(reason="Translate to keras.layers.add instead of math.add")
def test_translates_add(threepio):
    process_tests(add, threepio)


@pytest.mark.skip(reason="Translate to keras.layers.add instead of math.add")
def test_translates_d_add():
    process_tests(add2)


def test_translates_matmul(threepio):
    process_tests(matmul, threepio)


def test_translates_mean(threepio):
    process_tests(mean, threepio)


@pytest.mark.skip(reason="Requires mul -> multiply mapping")
def test_translates_mul(threepio):
    process_tests(mul, threepio)


@pytest.mark.skip(reason="Requires div -> truediv mapping")
def test_translates_div(threepio):
    process_tests(div, threepio)


@pytest.mark.skip(reason="Requires eq -> equal mapping")
def test_translates_eq(threepio):
    process_tests(eq, threepio)


def test_translates_sum(threepio):
    process_tests(torch_sum, threepio)


def test_translates_linear(threepio):
    for i, input_cmd in enumerate(linear["inputs"]):
        translations = threepio.translate(input_cmd, lookup_command=True)
        cmd_config = translations.pop(0)
        # Create a store for command outputs
        store = {}

        # Execute every translation
        for j, translation in enumerate(translations):
            result = translation.execute_routine(store)
            if translation.placeholder_output is not None:
                store[translation.placeholder_output] = result

        # Check end result
        answer = linear["answers"][i]
        check_answer(result, answer)


@pytest.mark.skip(reason="Requires dim -> axis mapping")
def test_translates_argmax(threepio):
    process_tests(argmax, threepio)


@pytest.mark.skip(reason="Requires t -> transpose mapping")
def test_translates_transpose(threepio):
    process_tests(t, threepio)


@pytest.mark.skip(reason="Requires sub -> subtract mapping")
def test_translates_subtract(threepio):
    process_tests(sub, threepio)


def test_translates_truediv(threepio):
    process_tests(truediv, threepio)



try:
    import importlib.resources as pkg_resources
except ImportError:
    import importlib_resources as pkg_resources
import json
import static


def get_mapped_commands() -> dict:
    json_txt = pkg_resources.read_text(static, "mapped_commands_full.json")
    return json.loads(json_txt)
